"""
Created on 09/07/2014

@author: synerty
"""
from twisted.internet import defer
from twisted.internet.defer import Deferred, DeferredList, succeed, inlineCallbacks
from vortex.DeferUtil import vortexLogFailure
from vortex.Payload import Payload
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.VortexFactory import VortexFactory

from peek_plugin_base.PeekVortexUtil import peekServerName

__author__ = "peek"

import logging

logger = logging.getLogger(__name__)

# -------------------------------------
# Software Update Handler for data from agents

# The filter we listen on
peekPlatformVersionFilt = {
    "plugin": "peak_platform",
    "key": "peek_platform.version.check",
}  # LISTEN / SEND


class PeekSwVersionPollHandler(object):
    PEEK_PLATFORM = "peek_platform"

    def __init__(self):
        self._startupDeferred = Deferred()
        self._ep = None

    def start(self):
        from peek_platform import PeekPlatformConfig

        if not PeekPlatformConfig.config.autoPackageUpdate:
            return succeed(True)

        self._ep = PayloadEndpoint(peekPlatformVersionFilt, self._process)

        d = VortexFactory.sendVortexMsg(
            vortexMsgs=Payload(filt=peekPlatformVersionFilt)
            .makePayloadEnvelope()
            .toVortexMsg(),
            destVortexName=peekServerName,
        )
        d.addErrback(vortexLogFailure, logger, consumeError=True)

        return self._startupDeferred

    def _process(self, payloadEnvelope: PayloadEnvelope, **kwargs):
        logger.info(payloadEnvelope.result)
        assert not payloadEnvelope.result  # result is None means success

        from peek_platform import PeekPlatformConfig

        platformUpdate = False
        deferredList = []

        for swVersionInfo in payloadEnvelope.tuples:
            if swVersionInfo.name == self.PEEK_PLATFORM:
                if PeekPlatformConfig.config.platformVersion != swVersionInfo.version:
                    logger.info(
                        "Recieved platform update new version is %s, we're %s",
                        swVersionInfo.version,
                        PeekPlatformConfig.config.platformVersion,
                    )
                    d = PeekPlatformConfig.peekSwInstallManager.update(
                        swVersionInfo.version
                    )
                    deferredList.append(d)

                    # Don't process any more, we'll update them when we restart.
                    platformUpdate = True
                    break

            else:
                installedPluginVer = PeekPlatformConfig.config.pluginVersion(
                    swVersionInfo.name
                )

                if installedPluginVer != swVersionInfo.version:
                    logger.info(
                        "Recieved %s update new version is %s, we're %s",
                        swVersionInfo.name,
                        swVersionInfo.version,
                        installedPluginVer,
                    )

                    d = PeekPlatformConfig.pluginSwInstallManager.update(
                        swVersionInfo.name, swVersionInfo.version
                    )
                    deferredList.append(d)

        if not self._startupDeferred:
            return

        if deferredList:
            if platformUpdate:
                self._startupDeferred.errback(
                    Exception("Startup stopped, Platform update and restart")
                )
            else:
                DeferredList(deferredList).chainDeferred(self._startupDeferred)

        else:
            logger.info("No update required")
            if self._startupDeferred:
                self._startupDeferred.callback(True)

        self._startupDeferred = None


peekSwVersionPollHandler = PeekSwVersionPollHandler()



import logging

from twisted.internet import protocol
from twisted.internet.defer import inlineCallbacks

from peek_platform.platform_init.init_platform import InitPlatform
from peek_plugin_base.PeekVortexUtil import peekServerName


logger = logging.getLogger("child_status_protocol")


class PluginSubprocChildStateProtocol(protocol.Protocol):
    COMMAND_LOAD = "LOAD"
    COMMAND_START = "START"
    COMMAND_STOP = "STOP"
    COMMAND_UNLOAD = "UNLOAD"
    COMMAND_SUCCESS = "SUCCESS"

    def __init__(self, serviceName: str, subprocessGroup: str):
        self._data = b""
        self._serviceName = serviceName
        self._subprocessGroup = subprocessGroup

    @inlineCallbacks
    def connectionMade(self):

        platformInitter = InitPlatform(
            self._serviceName, isPluginSubprocess=True
        )

        # Setup the platform
        platformInitter.setupPluginLoader()
        platformInitter.setupConfig()
        platformInitter.setupTwistedReactor()
        platformInitter.setupTempDirs()

        # Connect the vortex, only if we're not the logic service
        if self._serviceName != peekServerName:
            yield platformInitter.connectVortexClient()

    @inlineCallbacks
    def dataReceived(self, data: bytes):
        self._data += data

        while b"\n" in self._data:
            message, self._data = self._data.split(b"\n", 1)
            if not message:
                continue

            pluginName, command = message.decode().split(":", 1)

            yield self._runCommand(pluginName, command)

    @inlineCallbacks
    def _runCommand(self, pluginName: str, command: str):
        from peek_platform import PeekPlatformConfig

        try:

            if command == self.COMMAND_LOAD:
                yield PeekPlatformConfig.pluginLoader.loadStandalonePlugin(
                    pluginName
                )
            elif command == self.COMMAND_START:
                yield PeekPlatformConfig.pluginLoader.startStandalonePlugin(
                    pluginName
                )
            elif command == self.COMMAND_STOP:
                yield PeekPlatformConfig.pluginLoader.stopStandalonePlugin(
                    pluginName
                )
            elif command == self.COMMAND_UNLOAD:
                yield PeekPlatformConfig.pluginLoader.unloadStandalonePlugin(
                    pluginName
                )
            else:
                raise NotImplementedError(f"Unhandled command '{command}'")

            # Send the success response
            self.transport.write(
                f"{pluginName}:{self.COMMAND_SUCCESS}".encode()
            )
            self.transport.write(b"\n")

        except Exception as e:
            logger.exception(e)
            # Send the success response
            self.transport.write(
                (e.message if hasattr(e, "message") else str(e))
                .splitlines()[0]
                .encode()
            )
            self.transport.write(b"\n")



"""
 *
 *  Copyright Synerty Pty Ltd 2013
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
"""
import logging
import os
import sys
import tarfile
import urllib.error
import urllib.parse
import urllib.request
from abc import ABCMeta
from typing import Optional

from pytmpdir.directory_ import Directory
from twisted.internet import defer
from twisted.internet import reactor
from twisted.internet.defer import inlineCallbacks
from txhttputil.downloader.HttpFileDownloader import HttpFileDownloader
from txhttputil.util.DeferUtil import deferToThreadWrap

from peek_platform.WindowsPatch import isWindows
from peek_platform.util.PtyUtil import spawnPty, logSpawnException
from vortex.DeferUtil import deferToThreadWrapWithLogger

logger = logging.getLogger(__name__)

PEEK_PLATFORM_STAMP_FILE = "stamp"
"""Peek Platform Stamp File, The file within the release that conatins the version"""

IS_WIN_SVC = "isWinSvc"


class PeekSwInstallManagerABC(metaclass=ABCMeta):
    """Peek Software Install Manager ABC

    This class handles downloading the latest platform update from the server service
    installing it and then restarting this service.

    """

    def __init__(self):
        pass

    @classmethod
    def makeReleaseFileName(cls, version: str) -> str:
        """Make Release File Name

        This method creates an absolute path/filename for a peek release given it's
        stamp/version

        :param version: The stamp/version of the release
        :return: The absolute filename of the peek-release tar file
        """

        from peek_platform import PeekPlatformConfig

        return os.path.join(
            PeekPlatformConfig.config.platformSoftwarePath,
            "peek-release-%s.tar.gz" % version,
        )

    @classmethod
    def makePipArgs(cls, directory: Directory) -> [str]:
        """Make PIP Args

        This method creates the install arg list for pip, it's used both when testing
        when a new platform release is uploaded and the install on each service.

        :param directory: The directory where the peek-release is extracted to
        :return: The list of arguments to pass to pip
        """
        # Create an array of the package paths

        absFilePaths = [
            f.realPath
            for f in directory.files
            if f.name.endswith(".tar.gz") or f.name.endswith(".whl")
        ]

        # Create and return the pip args
        return [
            "install",  # Install the packages
            "--force-reinstall",  # Reinstall if they already exist
            "--no-cache-dir",  # Don't use the local pip cache
            "--no-index",  # Work offline, don't use pypi
            "--find-links",
            directory.path,
            # Look in the directory for dependencies
        ] + absFilePaths

    def notifyOfPlatformVersionUpdate(self, newVersion):
        self.installAndRestart(newVersion)

    @inlineCallbacks
    def update(self, targetVersion) -> Optional[str]:
        """Update

        This method is called when this service detects that the peek server has a newer
        version of software than this service.

        :param targetVersion: The target version to update to
        :return: The version that was updated to, or None if it failed
        """
        logger.info("Updating to %s", targetVersion)

        from peek_platform import PeekPlatformConfig

        url = (
            "http://%(ip)s:%(port)s/peek_logic_service.sw_install.platform.download?"
        ) % {
            "ip": PeekPlatformConfig.config.peekServerHost,
            "port": PeekPlatformConfig.config.peekServerPort,
        }

        args = {"name": PeekPlatformConfig.componentName}
        if targetVersion:
            args["version"] = str(targetVersion)

        url += urllib.parse.urlencode(args)

        try:
            file = yield HttpFileDownloader(url).run()

            if os.path.getsize(file.name) == 0:
                logger.warning(
                    "Peek server doesn't have any updates for %s, version %s",
                    PeekPlatformConfig.componentName,
                    targetVersion,
                )
                return

            yield self._installUpdate(targetVersion, file.name)

        except Exception as e:
            logger.exception(e)
            raise

        defer.returnValue(targetVersion)

    @inlineCallbacks
    def installAndRestart(self, targetVersion: str) -> None:
        newSoftwareTar = self.makeReleaseFileName(targetVersion)

        yield self._installUpdate(targetVersion, newSoftwareTar)

    @deferToThreadWrapWithLogger(logger)
    def _installUpdate(self, targetVersion: str, fullTarPath: str) -> str:
        """Install Update (Blocking)

        This method installs the packages in the latest peek-release.
        It then calls self.restartProcess to restart the service

        :param targetVersion: The version we should be updating to.
        :param fullTarPath: The path to the peek-release to install
        :return: The version that was installed, (from the file in the release)
        """

        from peek_platform import PeekPlatformConfig

        if not tarfile.is_tarfile(fullTarPath):
            raise Exception("Platform update download is not a tar file")

        directory = Directory()
        tarfile.open(fullTarPath).extractall(directory.path)
        directory.scan()

        stampFile = directory.getFile(name=PEEK_PLATFORM_STAMP_FILE)
        if not stampFile:
            raise Exception(
                "Peek release %s doesn't contain version stamp file %s"
                % (fullTarPath, PEEK_PLATFORM_STAMP_FILE)
            )

        with stampFile.open() as f:
            stampVersion = f.read().strip()

        if stampVersion != targetVersion:
            raise Exception(
                "Stamp file version %s doesn't match target version %s"
                % (stampVersion, targetVersion)
            )

        self._pipInstall(directory)

        PeekPlatformConfig.config.platformVersion = targetVersion

        # Call later, allow the server time to respond to the UI
        reactor.callLater(2.0, self.restartProcess)

        return targetVersion

    def _pipInstall(self, directory: Directory) -> None:
        """Pip Install

        Runs the PIP install for the packages provided in the directory

        :param directory: The directory containing the
        :return: None

        """

        pipExec = os.path.join(os.path.dirname(sys.executable), "pip")

        pipArgs = [sys.executable, pipExec] + self.makePipArgs(directory)

        # The platform update is tested for dependencies when it's first uploaded
        # PIP has a bug, when you have updated packages for several dependent files
        #   and try to install them all at once, some of the packages don't update.
        pipArgs += ["--no-deps"]

        pipArgs = " ".join(pipArgs)

        try:
            spawnPty(pipArgs)
            logger.info("Peek package update complete.")

        except Exception as e:
            logSpawnException(e)

            # Update the detail of the exception and raise it
            e.message = "Failed to install packages from the new release."
            raise

    # @abstractmethod
    # def _stopCode(self) -> None:
    #     """ Stop Code
    #
    #     This method should stop the running code, mainly timers and processing queues.
    #     Possibly web servers that may call database updates.
    #
    #     """
    #
    # @abstractmethod
    # def _updateCode(self):
    #     """ Update Code
    #
    #     This method is called to update any data, such as a database migration (typically)
    #     """
    #
    # @abstractmethod
    # def _startCode(self):
    #     """ Start Code
    #
    #     This is called when the update fails, the service should start back up and run as
    #     it did before the update.
    #
    #     """

    # def _synlinkTo(self, componentName, home, newPath):
    #     symLink = os.path.join(home, componentName)
    #     try:
    #         os.remove(symLink)
    #     except:
    #         pass
    #     os.symlink(newPath, symLink)

    def _restartProcessWinSvc(self) -> None:
        reactor.callFromThread(reactor.stop)

    def _restartProcessNormal(self) -> None:
        """Restart Process

        Restarts the current program.

        Note: this function does not return.
        Any cleanup action (like saving data) must be done before calling this function.

        Note: When peek is started by a windows service, this method is replaced with
        one that just restarts the windows service.

        """

        if IS_WIN_SVC in sys.argv:
            reactor.callFromThread(reactor.stop)
            return

        python = sys.executable
        argv = list(sys.argv)

        def addExe(val):
            if not "run_peek_" in val:
                return val
            if isWindows and not val.lower().endswith(".exe"):
                return val + ".exe"
            return val

        argv = map(addExe, argv)
        os.execl(python, python, *argv)

    restartProcess = (
        _restartProcessWinSvc
        if IS_WIN_SVC in sys.argv
        else _restartProcessNormal
    )


# run_peek_worker_service WILL NOT START if there are extra args
if IS_WIN_SVC in sys.argv:
    sys.argv.remove(IS_WIN_SVC)



"""
 *
 *  Copyright Synerty Pty Ltd 2013
 *
 *  This software is proprietary, you are not free to copy
 *  or redistribute this code in any format.
 *
 *  All rights to this software are reserved by 
 *  Synerty Pty Ltd
 *
 * Website : http://www.synerty.com
 * Support : support@synerty.com
 *
"""

import logging
import os
import shutil
from abc import ABCMeta

from jsoncfg.functions import save_config, ConfigWithWrapper

logger = logging.getLogger(__name__)

PEEK_LOGIC_SERVICE = "peek-logic-service"
PEEK_WORKER_SERVICE = "peek-worker-service"
PEEK_AGENT_SERVICE = "peek-agent-service"
PEEK_OFFICE_SERVICE = "peek-office-service"
PEEK_FIELD_SERVICE = "peek-field-service"
PEEK_SERVICES = (
    PEEK_LOGIC_SERVICE,
    PEEK_WORKER_SERVICE,
    PEEK_AGENT_SERVICE,
    PEEK_OFFICE_SERVICE,
    PEEK_FIELD_SERVICE,
)


class PeekFileConfigABC(metaclass=ABCMeta):
    """
    This class creates a basic agent configuration
    """

    DEFAULT_FILE_CHMOD = 0o600
    DEFAULT_DIR_CHMOD = 0o700

    __instance = None

    def __new__(cls):
        if cls.__instance is not None:
            return cls.__instance

        self = super(PeekFileConfigABC, cls).__new__(cls)
        cls.__instance = self
        return self

    def _migrateConfigDirv2tov3Move(self, oldServiceName):
        oldHomePath = os.path.expanduser("~/peek-%s.home" % oldServiceName)
        if not os.path.isdir(oldHomePath):
            return

        if not os.path.isdir(self._homePath):
            shutil.move(oldHomePath, self._homePath)

    def _migrateConfigDirv2tov3Copy(self, oldServiceName):
        oldHomePath = os.path.expanduser("~/peek-%s.home" % oldServiceName)
        if not os.path.isdir(oldHomePath):
            return

        if not os.path.isdir(self._homePath):
            shutil.copytree(oldHomePath, self._homePath)

    def _migrate(self):
        from peek_platform import PeekPlatformConfig

        copyServices = (PEEK_OFFICE_SERVICE, PEEK_FIELD_SERVICE)

        if PeekPlatformConfig.componentName == PEEK_LOGIC_SERVICE:
            self._migrateConfigDirv2tov3Move("server")

        elif PeekPlatformConfig.componentName == PEEK_WORKER_SERVICE:
            self._migrateConfigDirv2tov3Move("worker")

        elif PeekPlatformConfig.componentName == PEEK_AGENT_SERVICE:
            self._migrateConfigDirv2tov3Move("agent")

        elif PeekPlatformConfig.componentName in copyServices:
            self._migrateConfigDirv2tov3Copy("client")

    def __init__(self):
        """
        Constructor
        """
        from peek_platform import PeekPlatformConfig

        assert PeekPlatformConfig.componentName is not None

        self._homePath = os.path.join(
            os.path.expanduser("~"),
            "%s.home" % PeekPlatformConfig.componentName,
        )

        self._migrate()

        if not os.path.isdir(self._homePath):
            assert not os.path.exists(self._homePath)
            os.makedirs(self._homePath, self.DEFAULT_DIR_CHMOD)

        self._configFilePath = os.path.join(self._homePath, "config.json")

        if not os.path.isfile(self._configFilePath):
            assert not os.path.exists(self._configFilePath)
            with open(self._configFilePath, "w") as fobj:
                fobj.write("{}")

        self._cfg = ConfigWithWrapper(self._configFilePath)

        self._hp = "%(" + self._homePath + ")s"

    def _save(self):
        save_config(self._configFilePath, self._cfg)

    def _chkDir(self, path):
        if not os.path.isdir(path):
            assert not os.path.exists(path)
            os.makedirs(path, self.DEFAULT_DIR_CHMOD)
        return path



VORTEX_MSG_TO_CHILD_FD = 0
VORTEX_MSG_FROM_CHILD_FD = 1
LOGGING_FROM_CHILD_FD = 2
VORTEX_UUID_TO_CHILD_FD = 3
VORTEX_UUID_FROM_CHILD_FD = 4
PLUGIN_STATE_TO_CHILD_FD = 5
PLUGIN_STATE_FROM_CHILD_FD = 6



import gzip
import logging
import os
import sys
import zlib
from logging.handlers import RotatingFileHandler, SysLogHandler
from logging.handlers import TimedRotatingFileHandler

from pathlib import Path
from typing import Optional

LOG_FORMAT = "%(asctime)s %(levelname)s %(name)s:%(message)s"
DATE_FORMAT = "%d-%b-%Y %H:%M:%S"

logger = logging.getLogger(__name__)


def setupPeekLogger(serviceName: Optional[str] = None):
    logging.basicConfig(
        stream=sys.stdout,
        format=LOG_FORMAT,
        datefmt=DATE_FORMAT,
        level=logging.DEBUG,
    )

    if serviceName:
        updatePeekLoggerHandlers(serviceName)


def _namer(name):
    return name + ".gz"


def _rotator(source, dest):
    READ_CHUNK = 512 * 1024
    with open(source, "rb") as sf:
        with gzip.open(dest, "wb") as f:
            data = sf.read(READ_CHUNK)
            while data:
                f.write(data)
                data = sf.read(READ_CHUNK)
    os.remove(source)


def updatePeekLoggerHandlers(
    serviceName: Optional[str] = None, daysToKeep=28, logToStdout=True
):
    rootLogger = logging.getLogger()
    logFormatter = logging.Formatter(LOG_FORMAT, DATE_FORMAT)

    for handler in list(rootLogger.handlers):
        if isinstance(handler, TimedRotatingFileHandler):
            # Setup the file logging output
            rootLogger.removeHandler(handler)

        elif not sys.stdout.isatty() and not logToStdout:
            # Remove the stdout handler
            logger.info(
                "Logging to stdout disabled, see 'logToStdout' in config.json"
            )
            rootLogger.removeHandler(handler)

    fileName = str(Path.home() / ("%s.log" % serviceName))

    fh = TimedRotatingFileHandler(
        fileName, when="midnight", backupCount=daysToKeep
    )
    fh.setFormatter(logFormatter)
    fh.rotator = _rotator
    fh.namer = _namer
    rootLogger.addHandler(fh)


def setupLoggingToSyslogServer(host: str, port: int, facility: str):
    rootLogger = logging.getLogger()
    logFormatter = logging.Formatter(LOG_FORMAT, DATE_FORMAT)

    logging.getLogger().addHandler(logging.StreamHandler())

    if facility not in SysLogHandler.facility_names:
        logger.info(list(SysLogHandler.facility_names))
        raise Exception("Syslog facility name is a valid facility")

    facilityNum = SysLogHandler.facility_names[facility]

    fh = SysLogHandler(address=(host, port), facility=facilityNum)
    fh.setFormatter(logFormatter)
    rootLogger.addHandler(fh)



import logging
import os
import time
from abc import ABCMeta
import random
from typing import Optional

from jsoncfg.value_mappers import (
    require_string,
    RequireType,
    require_list,
    require_bool,
    require_integer,
)

from peek_platform.file_config.PeekFileConfigABC import PEEK_AGENT_SERVICE
from peek_platform.file_config.PeekFileConfigABC import PEEK_FIELD_SERVICE
from peek_platform.file_config.PeekFileConfigABC import PEEK_LOGIC_SERVICE
from peek_platform.file_config.PeekFileConfigABC import PEEK_OFFICE_SERVICE
from peek_platform.file_config.PeekFileConfigABC import PEEK_WORKER_SERVICE

logger = logging.getLogger(__name__)


class PeekFileConfigPlatformMixin(metaclass=ABCMeta):
    # --- Platform Logging

    @property
    def loggingDebugMemoryMask(self) -> int:
        with self._cfg as c:
            return c.logging.debugMemoryMask(0, require_integer)

    @property
    def loggingLevel(self) -> str:
        with self._cfg as c:
            lvl = c.logging.level("INFO", require_string)
            if lvl in logging._nameToLevel:
                return lvl

            logger.warning(
                "Logging level %s is not valid, defauling to INFO", lvl
            )
            return "INFO"

    @property
    def logToStdout(self) -> str:
        with self._cfg as c:
            return c.logging.logToStdout(False, require_bool)

    @property
    def daysToKeep(self) -> int:
        with self._cfg as c:
            val = c.logging.daysToKeep(14, require_integer)

            # As of v3.1+ cleanup the old log file properties
            for prop in ("rotateSizeMb", "rotationsToKeep"):
                if prop in c.logging:
                    logging = {}
                    logging.update(iter(c.logging))
                    del logging[prop]
                    c.logging = logging

            return val

    @property
    def loggingLogToSyslogHost(self) -> Optional[str]:
        with self._cfg as c:
            return c.logging.syslog.logToSysloyHost(None)

    @property
    def loggingLogToSyslogPort(self) -> int:
        with self._cfg as c:
            return c.logging.syslog.logToSysloyPort(514, require_integer)

    @property
    def loggingLogToSyslogFacility(self) -> str:
        with self._cfg as c:
            return c.logging.syslog.logToSysloyProtocol("user", require_string)

    @property
    def twistedThreadPoolSize(self) -> int:
        with self._cfg as c:
            count = c.twisted.threadPoolSize(500, require_integer)

        # Ensure the thread count is high
        if count < 50:
            logger.info("Upgrading thread count from %s to %s", count, 500)
            count = 500
            with self._cfg as c:
                c.twisted.threadPoolSize = count

        return count

    @property
    def autoPackageUpdate(self):
        with self._cfg as c:
            return c.platform.autoPackageUpdate(True, require_bool)

    # --- Platform Tmp Path
    @property
    def tmpPath(self):
        default = os.path.join(self._homePath, "tmp")
        with self._cfg as c:
            return self._chkDir(c.disk.tmp(default, require_string))

    # --- Platform Software Path
    @property
    def platformSoftwarePath(self):
        default = os.path.join(self._homePath, "platform_software")
        with self._cfg as c:
            return self._chkDir(
                c.platform.softwarePath(default, require_string)
            )

    # --- Platform Version
    @property
    def platformVersion(self):
        with self._cfg as c:
            return c.platform.version("0.0.0", require_string)

    @platformVersion.setter
    def platformVersion(self, value):
        with self._cfg as c:
            c.platform.version = value

    # --- Plugin Software Path
    @property
    def pluginSoftwarePath(self):
        default = os.path.join(self._homePath, "plugin_software")
        with self._cfg as c:
            return self._chkDir(c.plugin.softwarePath(default, require_string))

    # --- Plugin Data Path
    def pluginDataPath(self, pluginName):
        default = os.path.join(self._homePath, "plugin_data")

        with self._cfg as c:
            pluginData = c.plugin.dataPath(default, require_string)

        return self._chkDir(os.path.join(pluginData, pluginName))

    # --- Plugin Software Version
    def pluginVersion(self, pluginName):
        """Plugin Version

        The last version that we know about
        """
        with self._cfg as c:
            return c.plugin[pluginName].version(
                None, RequireType(type(None), str)
            )

    def setPluginVersion(self, pluginName, version):
        with self._cfg as c:
            c.plugin[pluginName].version = version

    # --- Plugins Installed
    @property
    def pluginsEnabled(self):
        with self._cfg as c:
            return c.plugin.enabled([], require_list)

    @pluginsEnabled.setter
    def pluginsEnabled(self, value):
        with self._cfg as c:
            c.plugin.enabled = value

    # --- Manhole
    @property
    def manholeEnabled(self) -> str:
        with self._cfg as c:
            return c.logging.manhole.enabled(True, require_bool)

    @property
    def manholePort(self) -> int:
        from peek_platform import PeekPlatformConfig

        port = {
            PEEK_LOGIC_SERVICE: 2201,
            PEEK_WORKER_SERVICE: 2202,
            PEEK_AGENT_SERVICE: 2203,
            PEEK_FIELD_SERVICE: 2204,
            PEEK_OFFICE_SERVICE: 2205,
        }[PeekPlatformConfig.componentName]
        with self._cfg as c:
            return c.logging.manhole.port(port, require_integer)

    @property
    def manholePassword(self) -> str:
        default = str(random.getrandbits(int(time.time() * 10**6 % 100000)))[
            :32
        ]
        with self._cfg as c:
            return c.logging.manhole.password(default, require_string)

    @property
    def manholePublicKeyFile(self) -> str:
        return self._ensureMaholeKeysExist()[0]

    @property
    def manholePrivateKeyFile(self) -> str:
        return self._ensureMaholeKeysExist()[1]

    def _ensureMaholeKeysExist(self) -> (str, str):
        with self._cfg as c:
            privateDefault = os.path.join(self._homePath, "manhole-key")
            priFile = c.logging.manhole.privateKeyFile(
                privateDefault, require_string
            )

            publicDefault = os.path.join(self._homePath, "manhole-key.pub")
            pubFile = c.logging.manhole.publicKeyFile(
                publicDefault, require_string
            )

            if not os.path.exists(priFile) or not os.path.exists(pubFile):
                self._manholeCreateKeys(priFile, pubFile)

            return pubFile, priFile

    def _manholeCreateKeys(self, priFile, pubFile):
        logger.info("(Re)Creating Manhole SSH Server Keys")

        from cryptography.hazmat.primitives import (
            serialization as crypto_serialization,
        )
        from cryptography.hazmat.primitives.asymmetric import rsa
        from cryptography.hazmat.backends import (
            default_backend as crypto_default_backend,
        )

        key = rsa.generate_private_key(
            backend=crypto_default_backend(),
            public_exponent=65537,
            key_size=2048,
        )
        privateKey = key.private_bytes(
            crypto_serialization.Encoding.PEM,
            crypto_serialization.PrivateFormat.TraditionalOpenSSL,
            crypto_serialization.NoEncryption(),
        ).decode()
        publicKey = (
            key.public_key()
            .public_bytes(
                crypto_serialization.Encoding.OpenSSH,
                crypto_serialization.PublicFormat.OpenSSH,
            )
            .decode()
        )

        from peek_platform import PeekPlatformConfig

        publicKey += " Peek %s Manhole" % PeekPlatformConfig.componentName

        with open(pubFile, "w") as f:
            f.write(publicKey)

        with open(priFile, "w") as f:
            f.write(privateKey)



from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType


@addTupleType
class PluginSubprocPlatformConfigTuple(Tuple):
    __tablename__ = "PluginSubprocPlatformConfigTuple"
    __tupleType__ = "peek_platform." + __tablename__

    serviceName: str = TupleField()
    subprocessGroup: str = TupleField()



import logging
import os
import subprocess
import sys

from typing import Optional

from vortex.DeferUtil import deferToThreadWrapWithLogger

from peek_platform.WindowsPatch import isWindows

logger = logging.getLogger(__name__)


class SpawnOsCommandException(subprocess.CalledProcessError):
    """Spawn OS Command Exception

    This exception is raised when an OS command fails or returns a non-zero exit code

    """

    def __init__(
        self,
        returncode: int,
        cmd: str,
        stdout: Optional[str] = None,
        stderr: Optional[str] = None,
        message: Optional[str] = None,
    ):
        """Constructor

        :param returncode: The code returned by the process.
        :param cmd: A string representing the command and args executed by bash.
        :param stdout: The output from the process.
        :param stderr: The error from the process.
        :param message: An additional message that can be used to emulate the message
        method of standard exceptions
        """

        subprocess.CalledProcessError.__init__(
            self, returncode, cmd, stdout, stderr
        )
        self.message = message

    def __str__(self):
        return "%s\nCommand '%s' returned non-zero exit status %d" % (
            self.message,
            self.cmd,
            self.returncode,
        )


class PtyOutParser:
    """PTY Out Parser

    The node tools require a tty, so we run it with

        parser = _PtyOutParser()
        import pty
        pty.spawn(\*args, parser.read)

    The only problem being that the output is sent to stdout, to solve this we intercept
    the output, return a . (dot) for every read, which it sends to stdout, and then only log
    the summary at the end of the webpack build.

    """

    def __init__(self, loggingStartMarker=None):
        """Constructor

        :param loggingStartMarker: If this is set, logging will not start until a line
        is starts with this marker.

        """
        self.data = ""
        self.startLogging = (
            False  # Ignore all the stuff before the final summary
        )
        self.allData = ""
        self.loggingStartMarker = loggingStartMarker

    def read(self, fd, size: Optional[int] = 1024):
        data = os.read(fd, size)
        self.data += data.decode(errors="ignore")
        self.allData += data.decode(errors="ignore")
        self.splitData()

        # Silence all the output
        if len(data):
            return b"."

        # If there is no output, return the EOF data
        return data

    def splitData(self):
        lines = self.data.splitlines(True)
        lines.reverse()
        while lines:
            line = lines.pop()
            if not line.endswith(os.linesep):
                self.data = line
                break
            self.logData(line.strip(os.linesep))

    def logData(self, line):
        self.startLogging = self.startLogging or (
            self.loggingStartMarker and line.startswith(self.loggingStartMarker)
        )

        if not (line and self.startLogging):
            return

        logger.debug(line)


def spawnPty(
    cmdAndArgs: str,
    parser: PtyOutParser = PtyOutParser(),
    env: Optional[dict] = None,
) -> None:
    """Spawn PTY

    This function spawns a PTY in Linux and falls back to using subprocess on windows.

    The commands are run via bash on both windows and posix compliant OSs. GitBash is
    recommended for windows.

    :param cmdAndArgs: A string containing the command and arguments to pass to bash.

    :param parser: An instance of c{PtyOutParser} that intercepts the PTY output.

    :returns: None, This either succeeds or raises an exception.

    :raises: c{CalledProcessError}
    """

    # If this is a posix compliant environment, then use pty
    if not isWindows:
        __reallySpawnSubprocess(cmdAndArgs, parser, env=env)
        return

    # Else, use subprocess
    spawnSubprocess(cmdAndArgs, env=env)


def __reallySpawnSubprocess(
    cmdAndArgs: str,
    parser: PtyOutParser = PtyOutParser(),
    env: Optional[dict] = None,
) -> None:
    from peek_platform import PeekPlatformConfig

    bashExec = PeekPlatformConfig.config.bashLocation

    logger.debug("Using interpreter : %s", bashExec)

    path, _, command = cmdAndArgs.partition("&&")
    path = path.strip("cd").strip()
    command = command.strip()

    import pexpect

    bashArgs = ["-l", "-c", command]

    logger.debug(
        f"Running command via pexpect : '{bashExec} {' '.join(bashArgs)}' in"
        f" folder '{path}'"
    )

    child = pexpect.spawn(
        bashExec,
        bashArgs,
        cwd=path,
        env=env,
        maxread=10000,  # a bigger value prints fewer dots
        use_poll=True,  # set to True when your `ulimit` is high
        encoding="utf-8",
    )

    class DotMaskedPtyLogFileWriter:
        def write(self, data):
            print(".", end="")

        def flush(self):
            sys.stdout.flush()

        def close(self):
            print("", flush=True)

    logfile = DotMaskedPtyLogFileWriter()
    child.logfile = logfile
    # alternate to show pty outputs
    # child.logfile = sys.stdout

    child.expect(pexpect.EOF, timeout=None)
    child.wait()
    logfile.close()
    child.close()

    logger.debug(f"subprocess returned status code {child.exitstatus}")
    if child.exitstatus:
        raise SpawnOsCommandException(
            child.exitstatus, cmdAndArgs, stdout="", stderr=parser.allData
        )


def spawnSubprocess(
    cmdAndArgs: str,
    parser: Optional[PtyOutParser] = None,
    env: Optional[dict] = None,
) -> None:
    """Spawn Subprocess

    This method calls an OS command using the subprocess package and bash as the
    interpreter.

    :param cmdAndArgs: A string containing the command and arguments to pass to bash.
    :param parser: The parser for the command output.

    :returns: None, This either succeeds or raises an exception.

    :raises: c{CalledProcessError}

    """
    from peek_platform import PeekPlatformConfig

    bashExec = PeekPlatformConfig.config.bashLocation

    logger.debug("Using interpreter : %s", bashExec)
    logger.debug("Running command via subprocess : %s", cmdAndArgs)

    commandComplete = subprocess.run(
        [cmdAndArgs],
        executable=bashExec,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,
        env=env,
    )

    if parser:
        parser.read(commandComplete.stdout, size=None)
        parser.read(commandComplete.stderr, size=None)

    if commandComplete.returncode:
        raise SpawnOsCommandException(
            commandComplete.returncode,
            cmdAndArgs,
            stdout=commandComplete.stdout.decode(),
            stderr=commandComplete.stderr.decode(),
        )

    return commandComplete


def logSpawnException(exception: Exception) -> None:
    """Log Exception

    This method logs the stderr / stdout data from an exception to the logger.

    :param exception: The exception to log the data from.
    """

    logger.exception(exception)

    if hasattr(exception, "message"):
        logger.error(exception.message)

    if hasattr(exception, "stderr"):
        [logger.error(l) for l in exception.stderr.splitlines()]

    if hasattr(exception, "stderr"):
        [logger.error(l) for l in exception.stderr.splitlines()]



import platform

from peek_platform.file_config.PeekFileConfigABC import PEEK_AGENT_SERVICE
from peek_platform.file_config.PeekFileConfigABC import PEEK_FIELD_SERVICE
from peek_platform.file_config.PeekFileConfigABC import PEEK_OFFICE_SERVICE
from peek_platform.file_config.PeekFileConfigABC import PEEK_WORKER_SERVICE

try:
    import win32serviceutil
    import win32service
    import win32event
    import win32api
except ImportError as e:
    if platform.system() is "Windows":
        raise

from twisted.internet import reactor
import time

import peek_platform
import logging

logger = logging.getLogger(__name__)


class PeekSvc(win32serviceutil.ServiceFramework):
    _svc_name_ = "peek-restarter"
    _svc_display_name_ = "Peek Restarter " + peek_platform.__version__
    _svc_deps_ = ["RpcSs"]

    def __init__(self, args):
        win32serviceutil.ServiceFramework.__init__(self, args)
        self.hWaitStop = win32event.CreateEvent(None, 0, 0, None)

    def SvcStop(self):
        self.ReportServiceStatus(win32service.SERVICE_STOP_PENDING)
        win32event.SetEvent(self.hWaitStop)

    def SvcDoRun(self):
        self.ReportServiceStatus(win32service.SERVICE_RUNNING)
        try:
            from peek_platform.util.LogUtil import setupPeekLogger

            setupPeekLogger(self._svc_name_)

            while True:
                retval = win32event.WaitForSingleObject(self.hWaitStop, 2000)
                if retval != win32event.WAIT_TIMEOUT:
                    break

                for service in (
                    PEEK_AGENT_SERVICE,
                    PEEK_WORKER_SERVICE,
                    PEEK_OFFICE_SERVICE,
                    PEEK_FIELD_SERVICE,
                ):
                    (
                        _,
                        status,
                        _,
                        errCode,
                        _,
                        _,
                        _,
                    ) = win32serviceutil.QueryServiceStatus(service)

                    if status != win32service.SERVICE_STOPPED:
                        continue

                    logger.info("Starting service %s", service)
                    win32serviceutil.StartService(service)
                    win32serviceutil.WaitForServiceStatus(
                        service, win32service.SERVICE_RUNNING, waitSecs=600
                    )
                    logger.info("Service %s started", service)

        except Exception as e:
            logger.exception(e)

        self.ReportServiceStatus(win32service.SERVICE_STOPPED)


def main():
    win32serviceutil.HandleCommandLine(PeekSvc)


if __name__ == "__main__":
    main()



import os
from abc import ABCMeta
from typing import Optional

from jsoncfg.value_mappers import require_bool
from jsoncfg.value_mappers import require_string, require_integer

from peek_platform.file_config.PeekFileConfigPlatformMixin import (
    PeekFileConfigPlatformMixin,
)
from peek_plugin_base.PeekPlatformServerInfoHookABC import (
    PeekPlatformServerInfoHookABC,
)


class PeekFileConfigDataExchangeClientMixin(PeekPlatformServerInfoHookABC):
    def __init__(self, config: PeekFileConfigPlatformMixin):
        self._config = config

    ### SERVER SECTION ###
    @property
    def peekServerHttpPort(self) -> int:
        with self._config._cfg as c:
            return c.dataExchange.httpPort(8011, require_integer)

    @property
    def peekServerHost(self) -> str:
        with self._config._cfg as c:
            return c.dataExchange.host("localhost", require_string)

    @property
    def peekServerSSL(self) -> bool:
        return self.peekServerUseSSL

    @property
    def peekServerUseSSL(self) -> bool:
        with self._config._cfg as c:
            return c.dataExchange.useSsl(False, require_bool)

    @property
    def peekServerSSLEnableMutualTLS(self) -> int:
        with self._config._cfg as c:
            return c.dataExchange.sslEnableMutualTLS(False, require_bool)

    @property
    def peekServerSSLClientBundleFilePath(self) -> Optional[str]:
        default = os.path.join(self._config._homePath, "key-cert-cachain.pem")
        with self._config._cfg as c:
            file = c.dataExchange.sslClientBundleFilePath(
                default, require_string
            )
            if os.path.exists(file):
                return file
            return None

    @property
    def peekServerSSLClientMutualTLSCertificateAuthorityBundleFilePath(
        self,
    ) -> Optional[str]:
        default = os.path.join(self._config._homePath, "root-cas.pem")
        with self._config._cfg as c:
            file = c.dataExchange.sslClientMutualTLSCertificateAuthorityBundleFilePath(
                default, require_string
            )
            if os.path.exists(file):
                return file
            return None

    @property
    def peekServerSSLMutualTLSTrustedPeerCertificateBundleFilePath(
        self,
    ) -> Optional[str]:
        default = os.path.join(self._config._homePath, "certs-of-peers.pem")
        with self._config._cfg as c:
            file = (
                c.dataExchange.sslMutualTLSTrustedPeerCertificateBundleFilePath(
                    default, require_string
                )
            )
            if os.path.exists(file):
                return file
            return None



import os
import shutil

from setuptools import find_packages
from setuptools import setup

from peek_platform.WindowsPatch import isWindows

###############################################################################
# Define variables
#
# Modify these values to fork a new plugin
#

author = "Synerty"
author_email = "contact@synerty.com"
py_package_name = "peek_platform"
pip_package_name = py_package_name.replace("_", "-")
package_version = "3.4.15"
description = "Peek Platform."

download_url = "https://bitbucket.org/synerty/%s/get/%s.zip"
download_url %= pip_package_name, package_version
url = "https://bitbucket.org/synerty/%s" % pip_package_name

###############################################################################

egg_info = "%s.egg-info" % pip_package_name
if os.path.isdir(egg_info):
    shutil.rmtree(egg_info)

if os.path.isfile("MANIFEST"):
    os.remove("MANIFEST")

# pip install flower==0.7.3 tornado==3.2.2


requirements = [
    # packages used for the platform to test and upgrade it's self
    "pip >= 21.0",
    "virtualenv >= 15.1.0",
    "twine",
    # networking and async framework. Peek is based on Twisted.
    "Cython >= 0.21.1",
    "Twisted[tls,conch]",
    "pyOpenSSL <= 19.1.0",  # PEEK-1136
    "pyasn1 >= 0.1.9",
    "pyasn1-modules >= 0.0.8",
    # Database
    "psycopg2-binary >= 2.7.6, < 2.9",  # PostGreSQL for Linux
    "SQLAlchemy < 1.4.0",  # Database abstraction layer
    "SQLAlchemy-Utils >= 0.32.9",
    "alembic >= 0.8.7",  # Database migration utility
    # installed and configured first
    # Utilities
    "python-dateutil >= 2.6.0",
    "Pygments >= 2.0.1",  # Generate HTML for code that is syntax styled
    "watchdog >= 0.8.3",
    # Used to detect file changes and re-copy them for frontend builds
    # Licensing
    "pycryptodome",
    "cryptography < 38.0.0",
    # Celery packages
    "flower<1.0.0",
    # "amqp >= 1.4.9",  # DEPENDENCY LINK BELOW
    # Potentially useful packages
    # "GitPython >= 2.0.8",
    # "jira",
    # "dxfgrabber >= 0.7.4",
    # Peek platform dependencies, all must match
    "peek-plugin-base",  ##==%s" % py_package_name,
    "peek-core-device",  ##==%s" % py_package_name,
    "peek-core-email",  ##==%s" % py_package_name,
    # Memory logging
    "psutil",
    # pty utility
    "pexpect>=4.8.0,<5.0",
]

lin_dependencies = [
    # We still require shapely on windows, but we need to manually download the win wheel
    "Shapely >= 1.5.16",  # Geospatial shape manipulation
    # We still require pymssql on windows, but we need to manually download the win wheel
    "pymssql",
    # Celery 4 is not supported on windows
    "future",  # This is required by celery
    "celery[redis,auth]<5.0.0",
]

win_dependencies = [
    # "pymssql >= 2.1.3",  # DB-API interface to Microsoft SQL Server, requires FreeTDS
    "pycparser >= 2.17",
    "cffi >= 1.9.1",
    "cryptography <= 3.1.1",  # PEEK-1136
    "pytest >= 3.0.5",
    "wheel >= 0.33.4.15",
    "virtualenv >= 15.1.0",
    # Celery 4 is not supported on windows
    "celery[redis,auth]<4.0.0",
    # Service support for windows
    "pypiwin32",
]

if isWindows:
    requirements.extend(win_dependencies)

else:
    requirements.extend(lin_dependencies)

# Packages that are presently installed from a git repo
# See http://stackoverflow.com/questions/17366784/setuptools-unable-to-use-link-from-dependency-links/17442663#17442663
dependency_links = [
    # Celery packages
    # "git+https://github.com/celery/py-amqp#egg=amqp",
]

dev_requirements = ["coverage >= 4.2", "mock >= 2.0.0", "selenium >= 2.53.6"]

requirements.extend(dev_requirements)

###############################################################################
# Define the dependencies

# Ensure the dependency is the same major number
# and no older then this version

# Force the dependencies to be the same branch
reqVer = ".".join(package_version.split(".")[0:2]) + ".*"

# >=2.0.*,>=2.0.6
requirements = [
    "%s==%s,>=%s" % (pkg, reqVer, package_version.split("+")[0])
    if pkg.startswith("peek")
    else pkg
    for pkg in requirements
]

###############################################################################
# Call the setuptools

setup(
    entry_points={
        "console_scripts": [
            "winsvc_peek_restarter = peek_platform.winsvc_peek_restarter:main"
        ]
    },
    name=pip_package_name,
    packages=find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests"]
    ),
    install_requires=requirements,
    zip_safe=False,
    version=package_version,
    description=description,
    author=author,
    author_email=author_email,
    url=url,
    download_url=download_url,
    keywords=["Peek", "Python", "Platform", "synerty"],
    classifiers=[],
)



import json
import logging
import os
import sys

from sqlalchemy.util import b64encode
from twisted.internet import reactor
from twisted.internet.defer import Deferred
from twisted.internet.defer import inlineCallbacks
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.Tuple import Tuple

from peek_platform.subproc_plugin_init.plugin_subproc import (
    plugin_subproc_child_main,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    LOGGING_FROM_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    PLUGIN_STATE_FROM_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    PLUGIN_STATE_TO_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    VORTEX_MSG_FROM_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    VORTEX_MSG_TO_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    VORTEX_UUID_FROM_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_constants import (
    VORTEX_UUID_TO_CHILD_FD,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_parent_protocol import (
    PluginSubprocParentProtocol,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_platform_config_tuple import (
    PluginSubprocPlatformConfigTuple,
)
from peek_platform.subproc_plugin_init.plugin_subproc.plugin_subproc_vortex_payload_envelope_tuple import (
    PluginSubprocVortexPayloadEnvelopeTuple,
)
from peek_plugin_base.PeekPlatformCommonHookABC import PeekPlatformCommonHookABC


logger = logging.getLogger(__name__)


class _NoKeycheckPayloadEndpoint(PayloadEndpoint):
    def _keyCheck(self, filt):
        pass


class PluginSubprocParentMain:
    def __init__(self, subprocessGroup: str):

        from peek_platform import PeekPlatformConfig

        self._serviceName = PeekPlatformConfig.componentName
        self._subprocessGroup = subprocessGroup

        self._processProtocol = PluginSubprocParentProtocol(
            self._subprocessGroup
        )

        platformConfigTupleEncoded = b64encode(
            json.dumps(
                PluginSubprocPlatformConfigTuple(
                    serviceName=self._serviceName,
                    subprocessGroup=self._subprocessGroup,
                ).toJsonDict()
            ).encode()
        )

        # Start the subprocess
        self._processTransport = reactor.spawnProcess(
            self._processProtocol,
            sys.executable,
            args=[
                sys.executable,
                plugin_subproc_child_main.__file__,
                platformConfigTupleEncoded,
            ],
            env=os.environ,
            path=os.path.dirname(plugin_subproc_child_main.__file__),
            childFDs={
                VORTEX_MSG_TO_CHILD_FD: "w",
                VORTEX_MSG_FROM_CHILD_FD: "r",
                LOGGING_FROM_CHILD_FD: "r",
                VORTEX_UUID_TO_CHILD_FD: "w",
                VORTEX_UUID_FROM_CHILD_FD: "r",
                PLUGIN_STATE_TO_CHILD_FD: "w",
                PLUGIN_STATE_FROM_CHILD_FD: "r",
            },
        )
        logger.debug("Spawned subprocess group %s", self._subprocessGroup)

    def __call__(
        self,
        pluginName: str,
        pluginRootDir: str,
        platform: PeekPlatformCommonHookABC,
    ):
        """This method simulates the call to the plugins constructor.

        :param pluginName:
        :param pluginRootDir:
        :param platform:
        :return:  PluginSubprocParentMainDelegate
        """
        from peek_platform.subproc_plugin_init.plugin_subproc_parent_main_delegate import (
            PluginSubprocParentMainDelegate,
        )

        return PluginSubprocParentMainDelegate(
            pluginName, pluginRootDir, platform, self
        )

    @inlineCallbacks
    def sendPayloadEnvelopeToChild(
        self,
        payloadEnvelope: PayloadEnvelope,
        vortexUuid: str,
        vortexName: str,
        **kwargs
    ):
        tuple_ = PluginSubprocVortexPayloadEnvelopeTuple(
            payloadEnvelope=payloadEnvelope,
            vortexUuid=vortexUuid,
            vortexName=vortexName,
        )
        encodedTuple = yield self._encodeTuple(tuple_)
        self._processTransport.write(encodedTuple)
        self._processTransport.write(b".")

    @deferToThreadWrapWithLogger(logger)
    def _encodeTuple(self, tuple_: Tuple) -> bytes:
        return b64encode(json.dumps(tuple_.toJsonDict()).encode()).encode()

    def sendPluginLoad(self, pluginName: str) -> Deferred:
        return self._processProtocol.sendPluginLoad(pluginName)

    def sendPluginStart(self, pluginName: str) -> Deferred:
        return self._processProtocol.sendPluginStart(pluginName)

    def sendPluginStop(self, pluginName: str) -> Deferred:
        return self._processProtocol.sendPluginStop(pluginName)

    def sendPluginUnload(self, pluginName: str) -> Deferred:
        return self._processProtocol.sendPluginUnload(pluginName)



from typing import Union

from vortex.PayloadEnvelope import VortexMsgList
from vortex.Tuple import Tuple
from vortex.Tuple import TupleField
from vortex.Tuple import addTupleType


@addTupleType
class PluginSubprocVortexMsgTuple(Tuple):
    __tablename__ = "PluginSubprocVortexMsgTuple"
    __tupleType__ = "peek_platform." + __tablename__

    # The vortexUuid to send the message to
    vortexUuid: str = TupleField()

    # Call arguments
    vortexMsgs: Union[VortexMsgList, bytes] = TupleField()

    priority: int = TupleField()



import logging
import os.path

from twisted.conch.checkers import InMemorySSHKeyDB
from twisted.conch.checkers import SSHPublicKeyChecker
from twisted.conch.manhole import ColoredManhole
from twisted.conch.manhole_ssh import ConchFactory
from twisted.conch.manhole_ssh import TerminalRealm
from twisted.conch.ssh.keys import Key
from twisted.cred.portal import Portal
from twisted.cred.checkers import InMemoryUsernamePasswordDatabaseDontUse
from twisted.internet import reactor

logger = logging.getLogger(__name__)


def start_manhole(
    port: int,
    manholeUserPassword: str,
    serverKeyPubPath: str,
    serverKeyPriPath: str,
):
    logger.info("Starting manhole server on port %s", port)

    def get_manhole(_):
        return ColoredManhole(globals())

    passwdChecker = InMemoryUsernamePasswordDatabaseDontUse(
        manhole=manholeUserPassword.encode()
    )

    checkers = [passwdChecker]

    usersPubKeyPath = os.path.expanduser("~/.ssh/id_rsa.pub")
    if os.path.exists(usersPubKeyPath):
        checkers.append(
            SSHPublicKeyChecker(
                InMemorySSHKeyDB({b"manhole": [Key.fromFile(usersPubKeyPath)]})
            )
        )

    realm = TerminalRealm()
    realm.chainedProtocolFactory.protocolFactory = get_manhole
    portal = Portal(realm, checkers=checkers)

    factory = ConchFactory(portal)
    factory.publicKeys[b"ssh-rsa"] = Key.fromFile(serverKeyPubPath)
    factory.privateKeys[b"ssh-rsa"] = Key.fromFile(serverKeyPriPath)

    reactorSocket = reactor.listenTCP(port, factory, interface="127.0.0.1")

    # Stop the manhole server when the reactor stops.4
    reactor.addSystemEventTrigger(
        "before", "shutdown", reactorSocket.stopListening
    )
